/* Democracy - a Bukkit plugin bringing voting features to Minecraft Servers
 * Copyright (c) 2011 Ed Halley <ed@halley.cc>
 * Offered under the MIT License.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

package cc.halley.bukkit.Democracy;

import cc.halley.kata.Filter;
import cc.halley.kata.S;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.logging.Logger;

import org.bukkit.ChatColor;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;
import org.bukkit.event.Event.Priority;
import org.bukkit.event.Event;
import org.bukkit.plugin.PluginDescriptionFile;
import org.bukkit.plugin.java.JavaPlugin;
import org.bukkit.plugin.PluginManager;
import org.bukkit.util.config.Configuration;

/**
 * Simple command-only plugin for Bukkit.
 * 
 * @author halley - Ed Halley
 */
public class Democracy
    extends JavaPlugin
{
	protected static Logger log;
	protected static Configuration config = null;

	protected Map<String,DemocracyIssue> issues =
		new HashMap<String,DemocracyIssue>();
	
	private final DemocracyPlayerListener playerListener =
	    new DemocracyPlayerListener(this);
	
	public void onEnable()
	{
		log = Logger.getLogger("Minecraft");
		config = getConfiguration();
		if (config.getInt("default.minimum", -1) <= 0)
		{
			log.warning("[Democracy] Writing config.yml with default values.");
			config.setProperty("basedir", "plugins/Democracy");
			config.setProperty("backend", "file");
			config.setProperty("backends.file.file", "proposals.data");
			//config.setProperty("backends.sqlite.file", "proposals.db");
			//config.setProperty("database.mysql.uri", "mysql://localhost/minecraft");
			//config.setProperty("database.mysql.user", "slartibartfast");
			//config.setProperty("database.mysql.pass", "not-important");
			config.setProperty("default.secrecy", false);
			config.setProperty("default.duration", "15m");
			config.setProperty("default.minimum", 5);
			//config.setProperty("voting.allow-nominate-offline", false);
			config.setProperty("limits.open-proposals", 5);
			config.setProperty("limits.open-proposals-per-player", 1);
			config.setProperty("voting.allow-nominating-self", false);
			config.setProperty("voting.proposing-includes-aye-vote", false);
			config.setProperty("voting.nominating-includes-aye-vote", true);
			config.setProperty("voting.allow-secret-vote", true);
			config.setProperty("voting.allow-public-vote", true);
			config.setProperty("voting.allow-change-vote", false);
			config.setProperty("voting.allow-vetoes", true);
			config.setProperty("voting.accountable-vetoes", true);
			config.setProperty("voting.allow-console-enact-veto", true);
			config.save();
		}

		// Register our events
		PluginManager pm = getServer().getPluginManager();
		pm.registerEvent(Event.Type.PLAYER_JOIN, playerListener,
		                 Priority.Normal, this);

		// Register our commands
		getCommand("vote").setExecutor(new DemocracyVoteCommand(this));
		getCommand("aye").setExecutor(new DemocracyVoteCommand(this));
		getCommand("nay").setExecutor(new DemocracyVoteCommand(this));
		getCommand("tally").setExecutor(new DemocracyTallyCommand(this));
		getCommand("propose").setExecutor(new DemocracyProposeCommand(this));
		getCommand("nominate").setExecutor(new DemocracyProposeCommand(this));
		getCommand("enact").setExecutor(new DemocracyEnactCommand(this));
		getCommand("veto").setExecutor(new DemocracyEnactCommand(this));

		loadIssues();
		
		scheduleIssueChecks();

		// Summarize the issues we loaded.
		int count = issues.size();
		log.info("[Democracy] Found " + count + " proposals.");
		if (count > 0)
			log.info("[Democracy] " + S.plural(count, "This", "These") +
			         " may be unopened, open for votes, " +
			         "or ready to tally.");

		// Let the admin know we're operational.
		PluginDescriptionFile pdfFile = this.getDescription();
		log.info(pdfFile.getName() +
		         " version " + pdfFile.getVersion() +
		         " is enabled!");
	}

	private void scheduleIssueChecks()
    {
		Runnable task = new Runnable() {
		    public void run() {
		    	Democracy.this.checkIssuesDue();
		    }
		};

		int first = 15 * 20;  /* seconds * 20 ticks/second */
		int period = 15 * 20; /* seconds * 20 ticks/second */
		getServer().getScheduler()
			.scheduleSyncRepeatingTask(this, task, first, period);		
    }

	protected void checkIssuesDue()
    {
		// Are we tapped out?
        if (issues.isEmpty())
        	return;

        // Does a tree fall in the forest?
        /*
        Player[] witnesses = getServer().getOnlinePlayers();
        if (witnesses == null || witnesses.length == 0)
        	return;
        */

        for (DemocracyIssue issue : issues.values())
        {
        	if (issue.isDue())
        	{
        		// Resolution happens automatically if possible.
        		issue.expire();
        		announceIssue(issue);

        		// One expiration per cycle, to reduce spam.
        		break;
        	}
        }
    }

	public void onDisable()
	{
		saveIssues();

		// Let the admin know we're down.
		PluginDescriptionFile pdfFile = this.getDescription();
		log.info(pdfFile.getName() + " has shut down.");
	}

	/*******************************************************************/

	@SuppressWarnings("unchecked")
    private void loadIssues()
    {
		log.info("[Democracy] Loading proposals...");

        String filename =
        	config.getString("basedir", ".") + File.separator +
        	config.getString("backends.file.file", "proposals.data");
        if (!new File(filename).exists())
        {
        	log.info("[Democracy] file " + filename + " not found.");
        	return;
        }
        try
        {
    		ObjectInputStream ois =
    			new ObjectInputStream(new FileInputStream(filename));
    		Object result = ois.readObject();
    		ois.close();
    		issues = (HashMap<String,DemocracyIssue>)result;
    		log.info("[Democracy] Load succeeded.");
    	}
        catch(Exception e)
        {
    		log.warning("[Democracy] Exception loading proposals: " +
    		            e.getClass().getName() + ": " + e.getMessage());
    		log.warning("[Democracy] Could not read or parse " + filename +
    		            "; all proposals voided.");
    		issues = new HashMap<String,DemocracyIssue>();
    	}
    }

	private void saveIssues()
    {
		log.info("[Democracy] Saving proposals...");

        String filename =
        	config.getString("basedir", ".") + File.separator +
        	config.getString("backends.file.file", "proposals.data");
        int count = issues.size();
        if (new File(filename).exists())
        {
        	log.info("[Democracy] Overwriting " + filename + " with " +
        	         count + S.plural(count, " proposal.", " proposals."));
        }
		try
		{
			ObjectOutputStream oos =
			    new ObjectOutputStream(new FileOutputStream(filename));
			oos.writeObject(issues);
			oos.flush();
			oos.close();
    		log.info("[Democracy] Save succeeded.");
		}
		catch (Exception e)
		{
    		log.warning("[Democracy] Exception saving proposals: " +
    		            e.getClass().getName() + ": " + e.getMessage());
    		log.warning("[Democracy] Could not write " + filename +
    		            "; proposals not saved.");
		}
    }

	/*******************************************************************/
	
	public void addIssue(DemocracyIssue issue)
		{ issues.put(issue.name, issue); }
	
	public void removeIssue(String name)
		{ if (issues.containsKey(name)) issues.remove(name); }
	
	public void removeIssue(DemocracyIssue issue)
		{ if (issue != null) removeIssue(issue.name); }

	public DemocracyIssue findIssue(String name)
    {
		if (issues.containsKey(name))
			return issues.get(name);
	    return null;
    }

	private static Filter<DemocracyIssue> _resolved =
	    new Filter<DemocracyIssue>() {
		    public boolean passes(DemocracyIssue product)
		    	{ return product.isClosed(); } };
	
	public Iterator<DemocracyIssue> getResolvedIssues()
	{
		return _resolved.filter(issues.values().iterator());
	}

	private static Filter<DemocracyIssue> _open =
	    new Filter<DemocracyIssue>() {
		    public boolean passes(DemocracyIssue issue)
		    	{ return issue.isOpened(); } };

	public Iterator<DemocracyIssue> getOpenIssues()
		{ return _open.filter(issues.values().iterator()); }

	private static Filter<DemocracyIssue> _unopened =
	    new Filter<DemocracyIssue>() {
		    public boolean passes(DemocracyIssue issue)
		    	{ return issue.isUnopened(); } };

	public Iterator<DemocracyIssue> getUnopenedIssues()
		{ return _unopened.filter(issues.values().iterator()); }

	public Iterator<DemocracyIssue> getAllIssues()
		{ return issues.values().iterator(); }
	
	/*******************************************************************/

	public void announceIssue(DemocracyIssue issue)
    {
		boolean resolved = issue.isClosed();
		boolean accountable = config.getBoolean("voting.accountable-vetoes",
		                                        true);
		String message = issue.summarizeStatus(null, accountable);

		for(Player player : getServer().getOnlinePlayers())
		{
			if (resolved && player.hasPermission("democracy.vote.tally"))
		    	player.sendMessage(ChatColor.GOLD + message);
			if (!resolved && player.hasPermission("democracy.vote.cast"))
		    	player.sendMessage(ChatColor.GOLD + message);
		}
		log.info("[Democracy] " + message);
    }

	public static void announceVote(DemocracyIssue issue, Player voter,
	                                String choice)
    {
		String who = voter.getDisplayName();
		String message =
			who + " voted '" + choice + "' " +
        	"in the '" + issue.name + "' proposal.";

		for(Player player : voter.getServer().getOnlinePlayers())
		{
			if (player.hasPermission("democracy.vote.tally") ||
				player.hasPermission("democracy.vote.cast"))
		    	player.sendMessage(ChatColor.GOLD + message);
		}
		log.info("[Democracy] " + message);
    }

	public void announceAction(DemocracyIssue issue, CommandSender sender,
	                           boolean enactment, String reason)
    {
		boolean accountable = config.getBoolean("voting.accountable-vetoes",
		                                        true);
		if (reason != null && !reason.isEmpty())
			reason = ": " + reason;

		String verb = "ratified";
		if (issue.resolution == null)
			verb = "cleared";
		else if (!enactment)
			verb = ChatColor.RED + "VETOED" + ChatColor.GOLD;

		if (issue.resolution == null)
			issue.resolution = "unresolved";
		else
			issue.resolution = "'" + issue.resolution + "'";

		String message;
		if (sender == null)
		{
			// The 'aye' vote in the 'throw-a-party' proposal has been
			// ratified: Mary's getting married.
			//
			message =
				"The " + issue.resolution + " vote in the '" +
				issue.name + "' proposal has been " + verb + reason + ".";
		}
		else
		{
			// Pow3rfiend has VETOED the 'pork' vote in the 'free-food'
			// proposal: no such thing as a free lunch, suckers.
			//
			if (sender instanceof Player)
				message = ((Player)sender).getDisplayName();
			else
				message = sender.getName();
			message =
				message + " has " + verb +
				" the " + issue.resolution + " vote in the '" +
				issue.name + "' proposal" + reason + ".";
		}

		if (enactment || accountable)
		{
    		for(Player player : getServer().getOnlinePlayers())
    		{
    			if (player.hasPermission("democracy.vote.tally") ||
    				player.hasPermission("democracy.vote.cast"))
    		    	player.sendMessage(ChatColor.GOLD + message);
    		}
		}
		log.info("[Democracy] " + ChatColor.stripColor(message));
    }
	
	/*******************************************************************/

	public Iterator<DemocracyIssue> getVotableIssues(final Player player)
    {
		Filter<DemocracyIssue> votable =
			new Filter<DemocracyIssue>() {
				public boolean passes(DemocracyIssue issue)
					{ return null ==
						issue.reasonVoteNotAllowed(player, null); } };
		return votable.filter(issues.values().iterator());
    }

	public int countVotableIssues(final Player player)
    {
		Filter<DemocracyIssue> votable =
			new Filter<DemocracyIssue>() {
				public boolean passes(DemocracyIssue issue)
					{ return null ==
						issue.reasonVoteNotAllowed(player, null); } };
		return votable.count(issues.values().iterator());
    }

	public int countProposedIssues(final Player player)
    {
		Filter<DemocracyIssue> proposed =
			new Filter<DemocracyIssue>() {
				public boolean passes(DemocracyIssue issue)
					{ return !issue.proponent.equals(player.getName()); } };
		return proposed.count(issues.values().iterator());
    }

	public int countResolvedIssues()
		{ return _resolved.count(issues.values().iterator()); }
	
	public int countAllIssues()
		{ return issues.size(); }
}
