/* Democracy - a Bukkit plugin bringing voting features to Minecraft Servers
 * Copyright (c) 2011 Ed Halley <ed@halley.cc>
 * Offered under the MIT License.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

package cc.halley.bukkit.Democracy;

import cc.halley.kata.S;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.bukkit.ChatColor;
import org.bukkit.entity.Player;

public class DemocracyIssue
	implements Serializable
{
    private static final long serialVersionUID = 8033407567007264117L;

	public final static int AYE = 0;
	public final static int CHOICE = 1;
	public final static int ROLLCALL = 2;
	public final static int NOMINATION = 3;
	public final static int PLAYERPICK = 4;

	public final static int PROPOSED = 0;
	public final static int OPENED = 1;
	public final static int EXPIRED = 2;
	public final static int RESOLVED = 3;
	public final static int ENACTED = 4;
	public final static int VETOED = 5;
	
	public String name;
	public int type = AYE;
	public int minimum = 1;           // required votes to succeed
	public long expires = -1;         // time span for votes, or closure time
	public boolean swing = false;     // allow vote-changing
	public boolean secret = true;     // don't announce votes globally
	public String nominee = null;     // the target of a /nominate
	public String proponent = null;   // the player who /propose'd
	public String description = null; // from the player who proposed
	public String resolution = null;  // the final winning choice, if clear
	public String vetoer = null;      // the player who /veto'd
	public String reason = null;      // from the player who vetoed

	// internal fields keep track of the votes
	private int total = 0;
	private int state = PROPOSED;
	private Map<String,String> votes = new HashMap<String,String>();
	private Map<String,Integer> tallies = new HashMap<String,Integer>();

	// automatic choice names in case of aye/nay type votes
	private Set<String> choices = null;
	private static final Set<String> aye_nay = new HashSet<String>(2);
	{
		if (aye_nay.isEmpty())
			{ aye_nay.add("aye"); aye_nay.add("nay"); }
	}
	
	// reserved words that would be confusing if used as an issue name
	private static final Set<String> reserved = new HashSet<String>(15);
	{
		String bad = "aye,nay,yes,no,veto,vote,tally,propose," +
			         "list,delete,cancel,help,0";
		for (String each : bad.split(","))
			reserved.add(each);
	}

	public DemocracyIssue(String name, String proponent)
	{
		this.name = name;
		this.proponent = proponent;
		choices = aye_nay;
	}

	public void setChoices(String[] candidates)
	{
		if (candidates == null)
		{
			choices = aye_nay;
			return;
		}
		choices = new HashSet<String>(candidates.length);
		for (String candidate : candidates)
			choices.add(candidate.trim().toLowerCase());
	}

	public static boolean isIssueNameReserved(String name)
	{
		if (name == null)
			return true;
		return reserved.contains(name.toLowerCase());
	}
	
	public boolean isVoteAllowed(Player player)
		{ return player.hasPermission("democracy.vote.cast"); }

	public boolean isUnopened()
		{ return (state == PROPOSED); }

	public void open()
	{
		if (state >= OPENED)
			throw new IllegalStateException();
		state = OPENED;
		expires += System.currentTimeMillis();
	}
	
	public boolean isOpened()
		{ return (state == OPENED); }

	public boolean isDue()
		{ return (state == OPENED && System.currentTimeMillis() >= expires); }

	public void expire()
	{
		if (state >= EXPIRED)
			throw new IllegalStateException();
		state = EXPIRED;
		expires = 0;
		attemptResolution();
	}

	public void attemptResolution()
	{
		if (state >= RESOLVED)
			throw new IllegalStateException();
		state = RESOLVED;
		List<String> leaders = rankVotersChoices();
		if (leaders == null)
			return;
		if (total < minimum)
			return;
		if (leaders.size() == 1)
		{
			resolution = leaders.get(0);
			return;
		}
	}

	public boolean isExpired()
	{
		if (state < EXPIRED)
			return false;
		long now = System.currentTimeMillis();
		if (expires >= 0 && now >= expires)
			state = EXPIRED;
		return (state == EXPIRED);
	}

	public boolean isClosed()
		{ return state >= EXPIRED; } // resolution != null || vetoer != null || isExpired(); }

	public boolean needsClosure()
		{ return isExpired() && state != ENACTED && state != VETOED; }

	public String getPlayerVote(Player player)
	{
		// get what they voted, or null for no vote on record
		String name = player.getName();
		if (!votes.containsKey(name))
			return null;
		return votes.get(name);
	}
	
	public String reasonVoteNotAllowed(Player player, String choice)
	{
		// not allowed? dismiss
		if (!isVoteAllowed(player))
			return "You are not allowed to vote on this proposal.";

		// vote closed? dismiss
		if (state != OPENED)
		{
			String reason = "";
			if (state == PROPOSED)
			{
				String who = player != null? player.getName() : "";
				if (who.equalsIgnoreCase(proponent))
					reason = " You need to open it.";
				else
					reason = " The author needs to open it.";
				// (we don't use 'proponent' in the reason
				//  because it may not be her displayname)
			}
			if (state == RESOLVED)
				reason = " The choice of '" + resolution + "' prevailed.";
			if (state == ENACTED)
				reason += " It was enacted.";
			if (state == VETOED)
				reason += " It was vetoed.";
			if (state == EXPIRED)
				reason += " It expired without a winner."; // tie or insufficent
			return "This proposal is not open for voting." + reason;
		}

		if (choice != null)
			choice = choice.toLowerCase().trim();

		// already voted? can't change it?
		String who = player.getName();
		if (votes.containsKey(who) && !swing)
		{
			// same choice as before? dismiss
			String prior = votes.get(who);
			if (choice != null && prior.equals(choice))
				return "You have already voted '" + prior + "'.";

			// not allowed to change? dismiss
			if (choice != null)
				return "You cannot change your vote " +
				                   "of '" + prior + "'.";
		}

		// not a valid choice? dismiss
		if (choice != null && !choices.contains(choice))
			return "You can only vote " + S.join(" or ", choices) + ".";
		
		return null;
	}

	public boolean castVote(Player player, String choice)
	{
		String who = player.getName();
		String reason = reasonVoteNotAllowed(player, choice);
		if (reason != null)
		{
			player.sendMessage(ChatColor.RED +
			                   reason);
			return false;
		}

		// subtract old vote if necessary
		if (votes.containsKey(who))
		{
		    // remove the old vote
			String old = votes.get(who);
			tallies.put(old, tallies.get(old)-1);
			votes.remove(who);
			total--;
		}

		// add the vote
		votes.put(who, choice);
		int was = tallies.containsKey(choice)? tallies.get(choice) : 0;
		tallies.put(choice, was+1);
		total++;
		
		if (secret)
		{
    		player.sendMessage(ChatColor.GREEN +
    		                   "You voted '" + choice + "' " +
    		                   "in the '" + name + "' proposal.");
		}
		else
		{
			Democracy.announceVote(this, player, choice);
		}
		return true;
	}

	private List<String> rankVotersChoices()
    {
		if (isUnopened())
			return null;
		if (tallies.size() == 0)
			return null;

		List<String> ranking = new ArrayList<String>(tallies.size());
		ranking.addAll(tallies.keySet());

		// a skunk election so far?
		if (tallies.size() == 1)
			return ranking;
		
		// at least two different choices so far...
		Collections.sort(ranking, new Comparator<String>() {
			public int compare(String a, String b) {
				return tallies.get(b).compareTo(tallies.get(a)); } });

		return ranking;
    }

	private void summarizeTitle(StringBuilder sb)
	{
		if (type == NOMINATION)
			sb.append("Nomination '");
		else
			sb.append("Proposal '");
		sb.append(name);
		sb.append("' ");
		if (description != null && !description.isEmpty())
		{
			sb.append("(");
			sb.append(description);
			sb.append(") ");
		}
	}
	
	public String summarizeResolution(boolean accountable)
    {
		StringBuilder sb = new StringBuilder();
		summarizeTitle(sb);
		if (isUnopened())
		{
			sb.append("has not opened for voting.");
			return sb.toString();
		}
		
		if (isOpened())
		{
			sb.append("is open for voting.");
			return sb.toString();
		}
		sb.append("has ");
		
		if (state == VETOED)
		{
			sb.append("been vetoed");
			if (accountable)
			{
				sb.append(" by ");
				sb.append(vetoer);
			}
			sb.append(".");
			return sb.toString();
		}

		if (state == EXPIRED && total < minimum)
		{
			sb.append("expired lacking votes.");
			return sb.toString();
		}

		if (resolution == null && state < RESOLVED)
			attemptResolution();
		// resolution will still be null if ties, etc.

		if (type == AYE || type == NOMINATION)
		{
			if ("aye".equals(resolution))
				sb.append("passed");
			else
				sb.append("failed");
		}
		else if (type == ROLLCALL)
		{
			int count = votes.size();
			sb.append("been joined by ");
			sb.append(count);
			sb.append(S.plural(count, " player", "players"));
		}
		else if (resolution != null)
		{
			sb.append("gone to '");
			sb.append(resolution);
			sb.append("'");
		}
		if (state >= EXPIRED && total < minimum)
			sb.append(" for insufficient votes");
		if (resolution != null && type != ROLLCALL)
		{
    		sb.append(" with ");
    		int count = tallies.get(resolution);
    		sb.append(count);
    		sb.append(" out of ");
    		sb.append(total);
    		sb.append(" votes (");
    		sb.append( 100 * count / total );
    		sb.append("%)");
		}
		
		// if (type != AYE) mention runner(s)-up?

		sb.append(".");
		return sb.toString();
    }

	public String summarizeChoices(String player)
    {
		StringBuilder sb = new StringBuilder();
		summarizeTitle(sb);
		if (player != null && votes.containsKey(player))
		{
			sb.append("received your vote of '");
			sb.append(votes.get(player));
			sb.append("'");
		}
		else
		{
    		sb.append("takes ");
    		if (type == AYE || type == NOMINATION || type == ROLLCALL)
    		{
    			sb.append("vote /aye "); sb.append(name);
    			sb.append(" or /nay "); sb.append(name);
    		}
    		else if (type == CHOICE)
    		{
    			sb.append("/vote "); sb.append(name);
    			sb.append(" ");
    			String sep = "";
    			for (String each : choices)
    			{
    				sb.append(sep); sep = " or ";
    				sb.append(each);
    			}
    		}
    		else
    		{
    			// unknown type
    			sb.append("prayer");
    		}
		}
		sb.append(".");
		return sb.toString();
    }

	public String summarizeStatus(String who, boolean accountable)
	{
		if (isUnopened())
			return summarizeUnopened();
		if (isClosed())
			return summarizeResolution(accountable);
		long remains = expires - System.currentTimeMillis();
		String remaining = "";
		if (remains >= 15*1000)
			remaining = " Now " + summarizeDuration(remains) + " remaining.";
		return summarizeChoices(who) + remaining;
	}

	public String summarizeUnopened()
    {
		StringBuilder sb = new StringBuilder();
		summarizeTitle(sb);
		sb.append("is not yet open; to allow votes, /");
		if (type == NOMINATION)
			sb.append("nominate ");
		else
			sb.append("propose ");
		sb.append(name);
		sb.append(" -open");
		return sb.toString();
    }

	public String summarizeOptions()
    {
		StringBuilder sb = new StringBuilder();
		summarizeTitle(sb);
		sb.append(secret?  "is secret," : "is public,");
		sb.append(" will run for ");
		sb.append(summarizeDuration(expires));
		sb.append(", and needs at least ");
		sb.append(minimum);
		sb.append(S.plural(minimum, " vote", " votes"));
		sb.append(".");
		return sb.toString();
    }

	public String summarizeTally()
    {
		StringBuilder sb = new StringBuilder();
		summarizeTitle(sb);
		sb.append("has ");
		int count = votes.size();
		if (type == ROLLCALL)
		{
			sb.append("been joined by ");
			sb.append(count);
			sb.append(S.plural(count, " player.", "players."));
			return sb.toString();
		}
		sb.append(count);
		sb.append(S.plural(count, " vote. ", " votes. "));

		List<String> leaders = rankVotersChoices();
		if (leaders == null && !isExpired())
			sb.append("It's no contest.");
		else if (leaders == null)
		{
		}
		else if (leaders.size() == 1)
		{
			sb.append("Choice '");
			sb.append(leaders.get(0));
			if (isExpired())
				sb.append("' swept.");
			else
				sb.append("' is sweeping.");
		}
		else
		{
    		int highest = tallies.get(leaders.get(0));
    		while (tallies.get(leaders.get(leaders.size()-1)) < highest)
    			leaders.remove(leaders.size()-1);
			if (leaders.size() == 1)
			{
				String only = leaders.get(0);
				count = tallies.get(only);
				sb.append("Choice '");
				sb.append(only);
				if (isExpired() && total < minimum)
					sb.append("' would have won but had ");
				else if (isExpired())
					sb.append("' won with ");
				else
					sb.append("' is winning with ");
				sb.append(count);
				sb.append(S.plural(count, " vote.", " votes."));
			}
			else
			{
				sb.append("Choices '");
				sb.append(S.join("' and '", leaders));
				if (isExpired())
					sb.append("' tied with ");
				else
					sb.append("' are tied with ");
				sb.append(highest);
				sb.append(S.plural(highest, " vote each.", " votes each."));
			}
		}
		if (total < minimum)
		{
			int need = minimum - total;
			if (isExpired())
				sb.append(" It needed ");
			else
				sb.append(" It needs ");
			sb.append(need);
			sb.append(" more ");
			sb.append(S.plural(need, "vote.", "votes."));
		}
		return sb.toString();
    }
	public String summarizeDuration(long amount)
    {
		long phrases = 2;
		long time = amount;
		time /= 1000L;
		StringBuilder sb = new StringBuilder();
		long n;
		n = time / (24*60*60);
		if (n > 0)
		{
			sb.append(n);
			sb.append(S.plural((int)n, " day", " days"));
			time -= n * 24*60*60;
			if (--phrases <= 0) return sb.toString();
			if (time > 0)
				sb.append(", ");
		}
		n = time / (60*60);
		if (n > 0)
		{
			sb.append(n);
			sb.append(S.plural((int)n, " hour", " hours"));
			time -= n * 60*60;
			if (--phrases <= 0) return sb.toString();
			if (time > 0)
				sb.append(", ");
		}
		n = time / 60;
		if (n > 0)
		{
			sb.append(n);
			sb.append(S.plural((int)n, " minute", " minutes"));
			time -= n * 60;
			if (--phrases <= 0) return sb.toString();
			if (time > 0)
				sb.append(", ");
		}
		if (time > 0)
		{
			n = time;
			sb.append(n);
			sb.append(S.plural((int)n, " second", " seconds"));
		}
		if (sb.length() == 0)
			sb.append("now");
	    return sb.toString();
    }
}
